/******************************************************************************
 * Arachnoid Graphics Plugin for Mupen64Plus
 * https://github.com/mupen64plus/mupen64plus-video-arachnoid/
 *
 * Copyright (C) 2009 Jon Ring
 * Copyright (C) 2007 Kristofer Karlsson, Rickard Niklasson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *****************************************************************************/

#ifndef CONFIG_H_
#define CONFIG_H_

#include "ConfigMap.h"
#include "m64p.h"
#include "m64p_types.h"

//Forward declaration
class GraphicsPlugin;

//*****************************************************************************
//* Config
//! Manages configuration of graphics plugin
//!
//! Responability:
//!  * Loads config data from core
//!  * Sets default config values
//!
//! @see ConfigMap
//*****************************************************************************
class Config
{
public:

    //Constructor / Destructor
    Config(GraphicsPlugin* graphicsPlugin);
    ~Config();

    bool initialize();

    //Loads settings from core
    void load();

    //Tell Graphics plugin that config has changed
    void updateGraphics();

public:

    //Get configuration
    ConfigMap* getConfig() { return &m_cfg; }

private:

    ConfigMap m_cfg;                     //!< Config map with all settings
    GraphicsPlugin* m_graphicsPlugin;    //!< Pointer to graphics plugin

    m64p_handle m_videoGeneralSection;
    m64p_handle m_videoArachnoidSection;
};

enum
{
    SCREEN_UPDATE_VI = 1,
    SCREEN_UPDATE_CI = 2
};

#endif
