/******************************************************************************
 * Arachnoid Graphics Plugin for Mupen64Plus
 * https://github.com/mupen64plus/mupen64plus-video-arachnoid/
 *
 * Copyright (C) 2007 Kristofer Karlsson, Rickard Niklasson
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 *****************************************************************************/

#ifndef N64_GAMES_H_
#define N64_GAMES_H_

//List of N64 Games
enum N64_ROM_ID
{
    UNKNOWN_ROM,
    _1080_SNOWBOARDING,
    BANJO_KAZOOIE,
    BANJO_TOOIE,
    BOMBERMAN_64,
    CONKERS_BAD_FUR_DAY,
    DIDDY_KONG_RACING,
    DONKEY_KONG_64,
    DR_MARIO,
    FIGHTERS_DESTINY,
    FIGHTER_DESTINY_2,
    F_ZERO_X,
    GOLDEN_EYE,
    LEGEND_OF_ZELDA_OCARINA_OF_TIME,
    MARIO_KART_64,
    PAPER_MARIO,
    PERFECT_DARK,
    SUPER_MARIO_64,
    SUPER_SMASH_BROS,
    STAR_FOX_64,        
    SOUTH_PARK_RALLY,    
    WAVE_RACE_64,
    WORLD_DRIVER_CHAMPOINSHIP,
    YOSHIS_STORY,
};


#endif
